/***********************************************************************
 * userfun.cpp: userdefined functions
 ***********************************************************************/
/*
  Defining new functions:

   - add the function name and argument count to RegisterFunctions(), e.g
       if (ifun == 0) {
         strcpy(cnam,"MyCtrl");
         *iVersion = 0;
       }
   
   - implement the function at the end of this source file
   
   - add the function name and ordinal number to the export file digexdyn.def, e.g.
       MyCtrl @2

*/

#include <stdio.h>
#include <math.h> 
#include <stdlib.h> 
#include <string.h> 
#include <ctype.h> 

#include "digexdyn.hpp"

#define ROUND(a) ( (a) > 0 ? (int(a+0.5)) : (int(a-0.5)) )


int __stdcall RegisterFunctions(int ifun, char* cnam, int* iVersion)
{ // register userdefined dynamic functions number ifun:
  //   return name (cnam[50]) and the actual version number (iVersion)
  //
  // return != 0, if ifun exceeds the number of available functions

  if (!cnam || !iVersion) return 1;

  if (ifun == 0) {
    strcpy(cnam,"MyCtrl");
    *iVersion = 0;
  
  } 
  else if (ifun == 1) {
    strcpy(cnam,"ShuntController");
    *iVersion = 0;
    
  } 
  else { // unknown function
    return 1;
  }

  return 0;
}



// *******************************************************************************************************************************************************************
//  Version 0:   
// *******************************************************************************************************************************************************************
//
//    void __cdecl function(double tEvent,double** dParams,const double** dOuts,const double** dIntSigs,char** eventstr,char* msg,int nParams,int nOuts,int nIntSigs)
//
//      Description:
//         double tEvent           : calling time in seconds (simulation time)
//         double** dParams        : parameter array (includes parameter and internal parameter, sorted by name)
//         const double** dOuts    : output signals array (interpolated values), values cannot be modified
//         const double** dIntSigs : not used (NULL), used for future extensions ...
//         char** eventstr         : char*[5]  max. five event strings can be define e.g. "create=EvtParam name=Test Target=this ...."
//                                             max. event string length char[100]
//         char* msg               : event message string to output a PCL message in the PowerFactory output window
//         int nParams             : no of parameters
//         int nOuts               : no of outputs
//         int nIntSigs            : not used, used for future extensions ...
//
// *******************************************************************************************************************************************************************


void __cdecl MyCtrl(double tEvent,double** dParams,const double** dOuts,const double** dIntSigs,char** eventstr,char* msg,int nParams,int nOuts,int nIntSigs)
{
  // strcpy(msg,"Event function called at:");
  sprintf(msg,"Event function called at: %f s %f hallo",(*dOuts)[0],(*dParams[0]));
}


void __cdecl ShuntController(double tEvent,double** dParams,const double** dOuts,const double** dIntSigs,char** eventstr,char* msg,int nParams,int nOuts,int nIntSigs)
{
  double *actSteps,*Stepslast;
  double Tdelay,Vbus,Qshunt,Vband,Vset;
  int Nsteps;

  actSteps = dParams[0]; 
  Nsteps = ROUND(*(dParams[1]));
  Qshunt = *(dParams[2]);
  Stepslast = dParams[3]; 
  Tdelay = *(dParams[4]);
  Vband = *(dParams[5]);
  Vset = *(dParams[6]);

  Vbus = *(dOuts[0]); // voltage at bus

  if (Vbus < (Vset-Vband)) { // step up -> increase tap nstep
    if ( (*actSteps)+1 < Nsteps) {
      (*actSteps) = (*actSteps) + 1;
    }
    else {
      sprintf(msg,"Could not increase capacitor. Max. steps reached.");
    }
  }
  else if (Vbus > (Vset+Vband)) { // step down -> decrease tap step
    if ( (*actSteps)-1 >= 0) {
      (*actSteps) = (*actSteps) - 1;
    }
    else {
      sprintf(msg,"Could not decrease capacitor. Min. steps reached.");
    }
  }

  if (fabs((*actSteps) - (*Stepslast)) > 0.1) { // no. of steps changed => set event 
    int newSteps=*actSteps;
    
    // put output message
    sprintf(msg,"Capacitor changed to Steps=%d",newSteps);
    
    // generator tap event for capacitance
    sprintf(eventstr[0],"create=EvtParam target=Shunt name=CapStep dtime=%f variable=ncapa value=%d",Tdelay,newSteps);
  }

  *Stepslast = *actSteps;
}